# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Fermionic state tests."""

from qiskit import QuantumCircuit
from qiskit.test import QiskitTestCase

from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom.fermions.fermion_circuit_solver import FermionicState


class TestFermionState(QiskitTestCase):
    """Class to test the fermion state class."""

    def test_initialize(self):
        """Test the initialization of fermionic states."""

        state = FermionicState([0, 1, 1, 0])

        self.assertEqual(state.sites, 4)
        self.assertEqual(state.num_species, 1)

        with self.assertRaises(QiskitColdAtomError):
            FermionicState([0, 2])

        with self.assertRaises(QiskitColdAtomError):
            FermionicState([[0, 1], [1, 0, 1]])

        state = FermionicState([[0, 1, 0], [1, 0, 1]])

        self.assertEqual(state.occupations_flat, [0, 1, 0, 1, 0, 1])
        self.assertEqual(state.num_species, 2)

    def test_string(self):
        """Test the string representation."""

        state = FermionicState([0, 1])

        self.assertEqual(str(state), "|0, 1>")

        state = FermionicState([[0, 1], [1, 0]])
        self.assertEqual(str(state), "|0, 1>|1, 0>")

    def test_occupations(self):
        """Test that to get the fermionic occupations."""
        state = FermionicState([0, 1])

        self.assertEqual(state.occupations, [[0, 1]])

    def test_from_flat_list(self):
        """Test the creation of fermionic states from flat lists."""

        state = FermionicState.from_total_occupations([0, 1, 1, 0], 2)
        self.assertEqual(state.occupations, [[0, 1], [1, 0]])

        with self.assertRaises(QiskitColdAtomError):
            FermionicState.from_total_occupations([0, 1, 1, 0], 3)

        state = FermionicState.from_total_occupations([0, 1, 1, 0], 1)
        self.assertEqual(state.occupations, [[0, 1, 1, 0]])

    def test_from_initial_state(self):
        """Test that we can load an initial state from a circuit."""

        circ = QuantumCircuit(4)
        circ.fload(0)
        circ.fload(2)

        state = FermionicState.initial_state(circ, 2)
        self.assertEqual(state.occupations, [[1, 0], [1, 0]])
        self.assertEqual(state.occupations_flat, [1, 0, 1, 0])

        state = FermionicState.initial_state(circ, 1)
        self.assertEqual(state.occupations, [[1, 0, 1, 0]])
        self.assertEqual(state.occupations_flat, [1, 0, 1, 0])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Backend to emulate fermionic tweezer experiments."""

from qiskit_cold_atom.fermions.fermion_simulator_backend import FermionSimulator


class FermionicTweezerSimulator(FermionSimulator):
    """Emulator backend of a fermionic device with n tweezer sites that uses two spin species.

    This backend will define a configuration with a set of supported gates and a line coupling map.
    At instantiation users can chose the length of the coupling map which by default is four sites.
    This backend is intended to be a realistic representation of a ferminoic tweezer experiment.
    For a general fermionic simulator backend, use the FermionSimulator base class.
    """

    def __init__(self, n_tweezers: int = 4, provider=None):
        """Create a new fermionic tweezer simulator backend.

        Args:
            n_tweezers: The number of optical tweezers.
            provider: The provider to which the backend may be added.
        """

        # define coupling maps for the gates
        sites = n_tweezers
        self._neighbouring_sites_couplings = [
            list(range(i, i + size)) + list(range(i + sites, i + sites + size))
            for size in range(2, sites + 1)
            for i in range(sites + 1 - size)
        ]
        self._global_site_couplings = [list(range(2 * sites))]
        self._single_site_couplings = [[i, i + sites] for i in range(sites)]

        configuration = {
            "backend_name": "fermionic_tweezer_simulator",
            "cold_atom_type": "fermion",
            "backend_version": "0.0.1",
            "simulator": True,
            "local": True,
            "coupling_map": None,
            "description": "Mock backend of a fermionic tweezer hardware. The first half of wires in a "
            "circuit denote the occupations of the spin-up fermions and the last half "
            "of wires denote the spin-down fermions",
            "basis_gates": [
                "fhop",
                "fint",
                "fphase",
                "fhubbard",
                "frx",
                "fry",
                "frz",
                "load",
            ],
            "num_species": 2,
            "memory": True,
            "n_qubits": 2 * sites,
            "conditional": False,
            "max_shots": 1e6,
            "max_experiments": 10,
            "open_pulse": False,
            "gates": [
                {
                    "name": "fhop",
                    "parameters": ["j_i"],
                    "qasm_def": "{}",
                    "description": "hopping of atoms to neighboring tweezers",
                    "coupling_map": self._neighbouring_sites_couplings,
                },
                {
                    "name": "fint",
                    "parameters": ["u"],
                    "qasm_def": "{}",
                    "description": "on-site interaction of atoms of opposite spin state",
                    "coupling_map": self._global_site_couplings,
                },
                {
                    "name": "fphase",
                    "parameters": ["mu_i"],
                    "qasm_def": "{}",
                    "description": "Applying a local phase to tweezers through an external potential",
                    "coupling_map": self._single_site_couplings,
                },
                {
                    "name": "frx",
                    "parameters": ["phi"],
                    "qasm_def": "{}",
                    "description": "x-rotation between the spin-up and spin-down state at one "
                    "tweezer site",
                    "coupling_map": self._single_site_couplings,
                },
                {
                    "name": "fry",
                    "parameters": ["phi"],
                    "qasm_def": "{}",
                    "description": "y-rotation between the spin-up and spin-down state at one "
                    "tweezer site",
                    "coupling_map": self._single_site_couplings,
                },
                {
                    "name": "frz",
                    "parameters": ["phi"],
                    "qasm_def": "{}",
                    "description": "z-rotation between the spin-up and spin-down state at one "
                    "tweezer site",
                    "coupling_map": self._single_site_couplings,
                },
            ],
            "supported_instructions": [
                "load",
                "measure",
                "barrier",
                "fhop",
                "fint",
                "fphase",
                "fhubbard",
                "frx",
                "fry",
                "frz",
            ],
        }

        super().__init__(config_dict=configuration, provider=provider)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Job for cold atom backends."""

from typing import Dict, Optional
import time
import json
import requests

from qiskit.providers import BackendV1 as Backend
from qiskit.providers import JobV1 as Job
from qiskit.providers import JobTimeoutError, JobError
from qiskit.providers import JobStatus
from qiskit.result import Result


class ColdAtomJob(Job):
    """Class of jobs returned by cold atom backends."""

    def __init__(self, backend: Backend, job_id: str):
        """
        Args:
            backend: The backend on which the job was run.
            job_id: The ID of the job.
        """
        super().__init__(backend, job_id)

        self.token = self._backend.token
        self.user = self._backend.username

    def _wait_for_result(self, timeout: float = None, wait: float = 5.0) -> Dict:
        """
        Query the backend to get the result.

        Args:
            timeout: time after which the server is no longer queried for the result
            wait: waiting time between queries for the result of the backend

        Returns:
            result dictionary formatted according to Qiskit schemas.

        Raises:
            JobTimeoutError: If the timeout is reached without receiving the result
            JobError: If the returned status reports an error
        """
        start_time = time.time()

        while True:
            elapsed = time.time() - start_time
            if timeout and elapsed >= timeout:
                raise JobTimeoutError("Timed out waiting for result")

            result = requests.get(
                self._backend.url + "/get_job_result",
                params={
                    "job_id": self._job_id,
                    "username": self.user,
                    "token": self.token,
                },
            ).json()

            if result["status"] == "finished":
                break
            if result["status"] == "error":
                raise JobError(f"{self._backend.name} returned error:\n" + str(result))
            time.sleep(wait)

        return result

    # pylint: disable=arguments-differ
    def result(self, timeout: Optional[float] = None, wait: float = 5.0) -> Result:
        """
        Retrieve a qiskit result object from the backend.

        Args:
            timeout: time after which the server is no longer queried for the result
            wait: waiting time between queries for the result of the backend

        Returns:
            qiskit.result object that contains the outcome of the job
        """
        result_dict = self._wait_for_result(timeout, wait=wait)

        return Result.from_dict(result_dict)

    def status(self):
        """
        Retrieve the status from the backend.

        Returns:
            status: A string describing the status of the job.
        """

        r = requests.get(
            self._backend.url + "/get_job_status",
            params={
                "job_id": self.job_id(),
                "username": self.user,
                "token": self.token,
            },
        )

        status_string = r.json()["status"]

        # If the backend can not be reached return ERROR as a status
        if r.status_code != 200:
            status = JobStatus.ERROR
        elif status_string == "INITIALIZING":
            status = JobStatus.INITIALIZING
        elif status_string == "QUEUED":
            status = JobStatus.QUEUED
        elif status_string == "VALIDATING":
            status = JobStatus.VALIDATING
        elif status_string == "RUNNING":
            status = JobStatus.RUNNING
        elif status_string == "CANCELLED":
            status = JobStatus.CANCELLED
        elif status_string == "DONE":
            status = JobStatus.DONE
        else:
            status = JobStatus.ERROR

        return status

    def error_message(self) -> Optional[str]:
        """
        Retrieve the error message from the backend.

        Returns:
            error: A string describing the error that happened on the backend.
        """
        status_payload = {"job_id": self.job_id()}

        r = requests.get(
            self._backend.url + "/get_job_status",
            params={
                "json": json.dumps(status_payload),
                "username": self.user,
                "token": self.token,
            },
        )

        return r.json().get("error_message", None)

    def cancel(self):
        raise NotImplementedError

    def submit(self):
        raise NotImplementedError



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Fermionic basis tests."""

from qiskit.test import QiskitTestCase
from qiskit_cold_atom.fermions.fermion_circuit_solver import (
    FermionicState,
    FermionicBasis,
)


class TestFermionicBasis(QiskitTestCase):
    """Class to test the fermionic basis class."""

    def test_initialize(self):
        """Test the initialization of fermionic basis."""
        with self.subTest("particle and spin conserved"):
            basis = FermionicBasis(sites=3, n_particles=[1, 2, 3])
            self.assertEqual(basis.num_species, 3)
            self.assertEqual(basis.dimension, 9)
        with self.subTest("particle number not conserved"):
            basis = FermionicBasis(sites=3, n_particles=[1, 2, 3], particle_conservation=False)
            self.assertEqual(basis.dimension, 512)
        with self.subTest("particle number not conserved"):
            basis = FermionicBasis(sites=3, n_particles=[1, 2, 3], spin_conservation=False)
            self.assertEqual(basis.dimension, 84)

    def test_string(self):
        """Test the string representation."""
        basis = FermionicBasis(sites=2, n_particles=[1, 1])
        expect = "\n 0.   |0, 1>|0, 1>\n 1.   |0, 1>|1, 0>\n 2.   |1, 0>|0, 1>\n 3.   |1, 0>|1, 0>"
        self.assertEqual(str(basis), expect)

    def test_init_from_state(self):
        """Test the initialization from a fermionic state"""

        state = FermionicState([[0, 1, 1], [1, 0, 0]])
        with self.subTest("particle and spin conserved"):
            basis = FermionicBasis.from_state(state, spin_conservation=True)
            self.assertEqual(basis.num_species, 2)
            self.assertEqual(basis.dimension, 9)
        with self.subTest("particle number not conserved"):
            basis = FermionicBasis.from_state(state, spin_conservation=False)
            self.assertEqual(basis.dimension, 20)
        with self.subTest("particle number not conserved"):
            basis = FermionicBasis.from_state(
                state, spin_conservation=False, particle_conservation=False
            )
            self.assertEqual(basis.dimension, 64)

    def test_get_occupations(self):
        """Test the representation as flat occupations lists."""

        basis = FermionicBasis(sites=2, n_particles=[1, 1])
        self.assertEqual(
            basis.get_occupations(),
            [[0, 1, 0, 1], [0, 1, 1, 0], [1, 0, 0, 1], [1, 0, 1, 0]],
        )

    def test_index_of_measurement(self):
        """Test the allocation of the basis state index for a given measurement."""
        basis = FermionicBasis(sites=2, n_particles=[1, 1])
        self.assertEqual(basis.get_index_of_measurement("0110"), 1)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Cold atom provider tests"""

import os
import uuid

from qiskit.test import QiskitTestCase
from qiskit_cold_atom.providers import ColdAtomProvider


class TestHeidelbergProvider(QiskitTestCase):
    """HeidelbergColdAtomProvider tests."""

    def setUp(self):
        super().setUp()

        # create directory for credentials if it doesn't already exist
        self._uuid = str(uuid.uuid4())

        self.path = os.path.join(os.path.expanduser("~"), ".qiskit" + self._uuid)
        self.path_exists = os.path.isdir(self.path)
        if not self.path_exists:
            os.mkdir(self.path)

        # Store some credentials
        self.username = "test_user"
        self.token = "test_token"
        self.url = "http://localhost:9000/shots"

        self.filename = os.path.join(self.path, "cold_atom_provider_test" + self._uuid)

    def test_credential_management(self):
        """Test the management of locally stored credential data"""
        with self.subTest("test save account"):
            ColdAtomProvider.save_account(
                url=self.url,
                username=self.username,
                token=self.token,
                filename=self.filename,
            )
            stored_credentials = ColdAtomProvider.stored_account(self.filename)
            target = {
                "urls": [self.url],
                "username": self.username,
                "token": self.token,
            }
            self.assertTrue(stored_credentials == target)

        with self.subTest("test overwrite warning"):
            with self.assertWarns(UserWarning):
                ColdAtomProvider.save_account(
                    url=self.url,
                    username=self.username,
                    token=self.token,
                    filename=self.filename,
                )

        with self.subTest("test add url"):
            ColdAtomProvider.add_url("second_url", filename=self.filename)
            stored_credentials = ColdAtomProvider.stored_account(self.filename)
            target = {
                "urls": [self.url, "second_url"],
                "username": self.username,
                "token": self.token,
            }
            self.assertTrue(stored_credentials == target)

        with self.subTest("test delete account"):
            with self.assertWarns(UserWarning):
                ColdAtomProvider.delete_account(
                    filename=os.path.join(os.path.expanduser("~"), ".qiskit", "wrong_filename")
                )
            ColdAtomProvider.delete_account(filename=self.filename)
            stored_credentials = ColdAtomProvider.stored_account(self.filename)
            self.assertTrue(stored_credentials == {})

    def test_provider_initialization(self):
        """Test the initialization of a cold atom provider from invalid credentials"""
        ColdAtomProvider.save_account(
            url=self.url,
            username=self.username,
            token=self.token,
            filename=self.filename,
        )
        with self.subTest("test load account"):
            # test that a warning is raised when initializing with invalid credentials
            with self.assertWarns(UserWarning):
                provider = ColdAtomProvider.load_account(filename=self.filename)
                target = {
                    "urls": [self.url],
                    "username": self.username,
                    "token": self.token,
                }
                self.assertTrue(provider.active_account() == target)

    def tearDown(self):
        """Clean-up after the test."""
        super().tearDown()

        if os.path.exists(self.filename):
            os.remove(self.filename)

        if not self.path_exists:
            os.rmdir(self.path)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module to describe a basis of fermionic states in occupation number representation."""

from typing import List, Union
from itertools import combinations, chain, product
import numpy as np

from qiskit_nature.operators.second_quantization import FermionicOp

from qiskit_cold_atom.fermions.fermionic_state import FermionicState


class FermionicBasis:
    """Class that represents the basis states of the fermionic fock space in occupation number
    representation for given particle numbers. The ordering of states complies with Qiskit bitstring
    ordering where the smaller number in binary representation has a lower index in the basis"""

    def __init__(
        self,
        sites: int,
        n_particles: Union[int, List[int]],
        particle_conservation: bool = True,
        spin_conservation: bool = True,
    ):
        """
        Args:
            sites: number of spatial fermionic modes
            n_particles: the total number of particles. If given as a list, the entries of the list
                give the particles per spin species, where the length of the list defines the number
                of different fermionic species
            particle_conservation: Boolean flag for the conservation of the total particle number
            spin_conservation: Boolean flag for conservation of the particle number per spin species
        """

        self.sites = sites

        if isinstance(n_particles, int):
            n_particles = [n_particles]

        self.n_particles = n_particles
        self.n_tot = sum(n_particles)
        self.num_species = len(n_particles)

        states = []

        if particle_conservation:
            if spin_conservation:

                indices = []
                for i, n in enumerate(self.n_particles):
                    indices.append(list(combinations(np.arange(sites) + i * sites, n)))

                for combination in product(*indices):
                    particle_indices = list(_ for _ in chain.from_iterable(combination))

                    occupations = [0] * self.sites * self.num_species
                    for idx in particle_indices:
                        occupations[idx] = 1

                    states.append(
                        FermionicState.from_total_occupations(occupations, self.num_species)
                    )

            else:
                for indices_tot in list(combinations(range(self.num_species * sites), self.n_tot)):
                    occupations_tot = [0] * sites * self.num_species
                    for i in indices_tot:
                        occupations_tot[i] = 1
                    states.append(
                        FermionicState.from_total_occupations(occupations_tot, self.num_species)
                    )
        else:
            for occs in product("10", repeat=(self.num_species * self.sites)):
                occupations_tot = [int(n) for n in occs]
                states.append(FermionicState(list(occupations_tot)))

        # reverse order of states to comply with Qiskit bitstring ordering
        self.states = states[::-1]

        self.dimension = len(self.states)

    def __str__(self):
        string = ""
        if self.dimension < 30:
            for i in range(self.dimension):
                if i < 10:
                    string += "\n {}.   ".format(i) + self.states[i].__str__()
                else:
                    string += "\n {}.  ".format(i) + self.states[i].__str__()
        else:
            for i in range(5):
                string += "\n {}.  ".format(i) + self.states[i].__str__()
            string += "\n . \n . \n ."
            for i in range(self.dimension - 5, self.dimension):
                string += "\n {}.  ".format(i) + self.states[i].__str__()

        return string

    @classmethod
    def from_state(
        cls,
        state: FermionicState,
        spin_conservation: bool,
        particle_conservation: bool = True,
    ):
        """Helper function to create the basis corresponding to a given occupation number state with
        particle number conservation and optionally spin conservation."""
        sites = state.sites
        n_particles = []
        for occs in state.occupations:
            n_particles.append(sum(occs))

        return cls(sites, n_particles, particle_conservation, spin_conservation)

    @classmethod
    def from_fermionic_op(cls, fer_op: FermionicOp):
        """Helper function to create the full Fock space basis corresponding to a given FermionicOp."""
        sites = fer_op.register_length
        n_particles = fer_op.register_length
        return cls(sites, n_particles, particle_conservation=False, spin_conservation=False)

    def get_occupations(self) -> List[List[int]]:
        """Get a list of the flattened occupations of the individual basis states."""
        return [state.occupations_flat for state in self.states]

    def get_index_of_measurement(self, bitstring: str) -> int:
        """
        For a binary string of occupations, e.g. '10100011', return the index of the basis state
        that corresponds to these occupations. In contrast to qubits, this index is not given by 2
        to the power of the bitstring as fermionic bases do not always include all states due to
        particle and spin conservation rules.

        Args:
            bitstring: A binary string of occupations.

        Returns:
            The index of the basis state corresponding to the bitstring.
        """
        occupation_strings = ["".join(map(str, k)) for k in self.get_occupations()]
        index = occupation_strings.index(bitstring)
        return index



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Module to support spin circuits.

The spins module holds the circuit instructions and simulators needed for cold-atom setups which
control the collective spin of an ensemble of atoms in a Bose-Einstein-condensate.

In this setting, each wire in a quantum circuit describes a single quantum mechanical angular momentum
(or spin) with principal quantum number :math:`S`. Upon measurement, each wire can be found in a state
ranging from :math:`0` to :math:`2S`.
Backends to describe such spin circuits are subclasses of the :class:`BaseSpinBackend` class.
The :class:`SpinSimulator` backend is a general purpose simulator backend that simulates spin
circuits.

Spin backends
-------------------
.. autosummary::
   :toctree: ../stubs/

   BaseSpinBackend
   SpinSimulator

At the start of the circuit each spin is taken to be initialized in the 0 state, in analogy to qubits.


Spin gates
==========

Spin gates are quantum circuit instructions designed specifically for cold-atom based setups that
control large spin ensembles. These gates are characterized by their effect on the spin ensemble.
Spin gates are subclasses or instances of the :class:`SpinGate` class. All of these gates define a
``generator`` property used to compute the time-evolution. These generators are second quantized
operators (:class:`SpinOp` from Qiskit Nature) that describe the gate Hamiltonian acting on the spins.
When an entry of the :mod:`qiskit_cold_atom.spins` module is imported, the Spin gates are added to the
:class:`QuantumCircuit` class in Qiskit.

The module includes a number of gates suitable to a platform that implements rotations and squeezing of
collective spins.

.. autosummary::
   :toctree: ../stubs/

    SpinGate
    RLXGate
    RLYGate
    RLZGate
    RLZ2Gate
    OATGate
    RLZLZGate
    RLXLYGate

These gates should serve as an example of how a concrete collective spin platform can be described
through :mod:`qiskit_cold_atom`.
Users are encouraged to define their own gates to describe different collective spin experiments.
If these gates define a ``generator`` property as laid out above, the :class:SpinSimulator
can be used to simulate circuits with such custom gates.


Circuit solvers
===============
Circuit solvers are classes that allow users to simulate a quantum circuit
for cold-atom based setups. They are subclasses of
:class:`BaseCircuitSolver` and can be called on quantum circuits to solve them.
The numerical simulation of spin circuits is carried out by the :class:`SpinCircuitSolver` class.
This simulates the circuits via exact diagonalization and provides access to the unitary,
the statevector and simulated measurement outcomes of the circuit.

.. autosummary::
   :toctree: ../stubs/

   SpinCircuitSolver
"""

from .spin_simulator_backend import SpinSimulator
from .base_spin_backend import BaseSpinBackend
from .spin_circuit_solver import SpinCircuitSolver

# Gate imports
from .spins_gate_library import (
    SpinGate,
    RLXGate,
    RLYGate,
    RLZGate,
    RLZ2Gate,
    OATGate,
    RLZLZGate,
    RLXLYGate,
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2018, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import sys
import setuptools
import inspect

long_description = """Qiskit cold atom is an open-source library to describe cold atomic quantum 
experiments in a gate- and circuit-based framework."""

with open("requirements.txt") as f:
    REQUIREMENTS = f.read().splitlines()

VERSION_PATH = os.path.join(os.path.dirname(__file__), "qiskit_cold_atom", "VERSION.txt")
with open(VERSION_PATH, "r") as version_file:
    VERSION = version_file.read().strip()

setuptools.setup(
    name="qiskit-cold-atom",
    version=VERSION,
    description="Integration of cold atomic experiments into the Qiskit SDK.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/qiskit-community/qiskit-cold-atom",
    author="Qiskit cold atom development team",
    license="Apache 2.0",
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
    keywords="qiskit sdk quantum cold atoms",
    packages=setuptools.find_packages(include=["qiskit_cold_atom", "qiskit_cold_atom.*"]),
    install_requires=REQUIREMENTS,
    include_package_data=True,
    python_requires=">=3.6",
    zip_safe=False,
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Fermionic gate tests """

import numpy as np
from scipy.linalg import expm
from qiskit.test import QiskitTestCase
from qiskit_nature.operators.second_quantization import FermionicOp
from qiskit_cold_atom.fermions.fermion_simulator_backend import FermionSimulator
from qiskit_cold_atom.fermions.fermion_gate_library import (
    Hop,
    Interaction,
    Phase,
    FRXGate,
    FRYGate,
    FRZGate,
    FermiHubbard,
)


class TestFermionicGates(QiskitTestCase):
    """Tests for the fermionic gates"""

    def setUp(self):
        super().setUp()
        self.backend = FermionSimulator()

    def test_interaction_gate(self):
        """check matrix form of interaction gate in a two-tweezer example"""
        u_val = np.pi / 2
        circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ.append(Interaction(4, u_val), qargs=[0, 1, 2, 3])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ_decorated.fint(u_val, [0, 1, 2, 3])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, num_species=2).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.array(
                            [
                                [u_val, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, u_val],
                            ]
                        )
                    ),
                )
            )

    def test_hopping_gate(self):
        """check matrix form of hopping gate in a two-tweezer example"""
        j = np.pi / 4
        circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ.append(Hop(4, [j]), qargs=[0, 1, 2, 3])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ_decorated.fhop([j], [0, 1, 2, 3])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, num_species=2).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.array(
                            [
                                [0, -j, -j, 0],
                                [-j, 0, 0, -j],
                                [-j, 0, 0, -j],
                                [0, -j, -j, 0],
                            ]
                        )
                    ),
                )
            )

    def test_phase_gate(self):
        """check matrix form of phase gate in a two-tweezer example"""
        phi1, phi2 = np.pi / 4, np.pi / 8
        circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ.append(Phase(4, [phi1, phi2]), qargs=[0, 1, 2, 3])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ_decorated.fphase([phi1, phi2], [0, 1, 2, 3])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, num_species=2).result().get_unitary()

            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.array(
                            [
                                [2 * phi2, 0, 0, 0],
                                [0, phi2 + phi1, 0, 0],
                                [0, 0, phi2 + phi1, 0],
                                [0, 0, 0, 2 * phi1],
                            ]
                        )
                    ),
                )
            )

    def test_spin_rx_gate(self):
        """check matrix form of spin_rx gate in a two-tweezer example"""
        phi = np.pi / 4
        circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ.append(FRXGate(phi), qargs=[0, 2])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ_decorated.frx(phi, [0, 2])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, num_species=2).result().get_unitary()

            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.array(
                            [
                                [0, 0, 0, phi, 0, 0],
                                [0, 0, 0, 0, 0, 0],
                                [0, 0, 0, 0, 0, -phi],
                                [phi, 0, 0, 0, 0, 0],
                                [0, 0, 0, 0, 0, 0],
                                [0, 0, -phi, 0, 0, 0],
                            ]
                        )
                    ),
                )
            )

    def test_spin_ry_gate(self):
        """check matrix form of spin_ry gate in a two-tweezer example"""
        phi = np.pi / 4
        circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ.append(FRYGate(phi), qargs=[0, 2])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ_decorated.fry(phi, [0, 2])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, num_species=2).result().get_unitary()

            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.array(
                            [
                                [0, 0, 0, 1j * phi, 0, 0],
                                [0, 0, 0, 0, 0, 0],
                                [0, 0, 0, 0, 0, -1j * phi],
                                [-1j * phi, 0, 0, 0, 0, 0],
                                [0, 0, 0, 0, 0, 0],
                                [0, 0, 1j * phi, 0, 0, 0],
                            ]
                        )
                    ),
                )
            )

    def test_spin_rz_gate(self):
        """check matrix form of spin_rz gate in a two-tweezer example"""
        phi = np.pi / 4
        circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ.append(FRZGate(phi), qargs=[0, 2])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ_decorated.frz(phi, [0, 2])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, num_species=2).result().get_unitary()

            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.array(
                            [
                                [0, 0, 0, 0],
                                [0, -phi, 0, 0],
                                [0, 0, phi, 0],
                                [0, 0, 0, 0],
                            ]
                        )
                    ),
                )
            )

    def test_fermionic_gate_class(self):
        """test the functionality of the base class for fermionic gates"""

        test_gates = [
            Hop(num_modes=4, j=[0.5]),
            Interaction(num_modes=8, u=2.0),
            Phase(num_modes=2, mu=[1.0]),
            FRXGate(phi=0.5),
            FRYGate(phi=-0.7),
            FRZGate(phi=1.3),
            FermiHubbard(num_modes=4, j=[0.5], u=1.0, mu=[0.4, 1.2]),
        ]

        with self.subTest("test to_matrix and power"):
            for gate in test_gates:
                exp_matrix = gate.to_matrix() @ gate.to_matrix()
                exp_gate = gate.power(2)
                self.assertTrue(np.allclose(exp_matrix, exp_gate.to_matrix()))

        with self.subTest("test generation of operator"):
            from qiskit.quantum_info import Operator

            for gate in test_gates:
                self.assertTrue(isinstance(Operator(gate), Operator))

    def test_identity_gates(self):
        """test that gates with parameters equal to zero still have a well-defined generator."""
        test_gates = [
            Hop(num_modes=4, j=[0.0]),
            Interaction(num_modes=4, u=0.0),
            Phase(num_modes=2, mu=[0.0]),
            FRXGate(phi=0.0),
            FRYGate(phi=-0.0),
            FRZGate(phi=0.0),
            FermiHubbard(num_modes=4, j=[0.0], u=0.0, mu=[0.0, 0.0]),
        ]

        for gate in test_gates:
            self.assertIsInstance(gate.generator, FermionicOp)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""General spin simulator backend."""

from typing import Union, List, Dict, Any, Optional
import uuid
from fractions import Fraction
import warnings
import time
import datetime

from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options
from qiskit import QuantumCircuit
from qiskit.result import Result
from qiskit.circuit.measure import Measure
from qiskit_aer import AerJob

from qiskit_cold_atom.spins.spin_circuit_solver import SpinCircuitSolver
from qiskit_cold_atom.spins.base_spin_backend import BaseSpinBackend
from qiskit_cold_atom.circuit_tools import CircuitTools


class SpinSimulator(BaseSpinBackend):
    """A simulator to simulate general spin circuits.

    This general spin simulator backend simulates spin circuits with gates that have
    generators described by spin Hamiltonians. It computes the statevector and unitary
    of a circuit and simulates measurements.
    """

    # Default configuration of the backend if the user does not provide one.
    __DEFAULT_CONFIGURATION__ = {
        "backend_name": "spin_simulator",
        "backend_version": "0.0.1",
        "n_qubits": None,
        "basis_gates": None,
        "gates": [],
        "local": True,
        "simulator": True,
        "conditional": False,
        "open_pulse": False,
        "memory": True,
        "max_shots": 1e5,
        "coupling_map": None,
        "description": "a base simulator for spin circuits. Instead of a qubit, each wire represents a "
        "single high-dimensional spin",
    }

    def __init__(self, config_dict: Optional[Dict[str, Any]] = None, provider=None):
        """
        Initialize the backend from a configuration dictionary.

        Args:
            config_dict: Configuration dictionary of the backend. If None is given
                a default is assumed.
        """

        if config_dict is None:
            config_dict = self.__DEFAULT_CONFIGURATION__

        super().__init__(
            configuration=BackendConfiguration.from_dict(config_dict), provider=provider
        )

    @classmethod
    def _default_options(cls):
        return Options(shots=1)

    def _execute(self, data: Dict[str, Any], job_id: str = "") -> Result:
        """
        Helper function to execute a job. The circuit and all relevant parameters are
        given in the data dict. Performs validation checks on the received circuits
        and utilizes the SpinCircuitSolver to perform the numerical simulations.

        Args:
            data: Data dictionary that that contains the experiments to simulate, given in the shape:
                data = {
                    "num_species": int,
                    "shots": int,
                    "seed": int,
                    "experiments": Dict[str, QuantumCircuit],
                }
            job_id: The job id assigned by the run method

        Returns:
            result: A qiskit job result.
        """
        # Start timer
        start = time.time()

        output = {"results": []}

        spin = data["spin"]
        shots = data["shots"]
        seed = data["seed"]

        solver = SpinCircuitSolver(spin, shots, seed)

        for exp_i, exp_name in enumerate(data["experiments"]):
            experiment = data["experiments"][exp_name]
            circuit = experiment["circuit"]

            # perform compatibility checks with the backend configuration in case gates and supported
            # instructions are constrained by the backend's configuration
            if self.configuration().gates and self.configuration().supported_instructions:
                CircuitTools.validate_circuits(circuits=circuit, backend=self, shots=shots)

            # check whether all wires are measured
            measured_wires = set()

            for inst in circuit.data:
                if isinstance(inst[0], Measure):
                    for wire in inst[1]:
                        index = circuit.qubits.index(wire)
                        if index in measured_wires:
                            warnings.warn(
                                f"Wire {index} has already been measured, "
                                f"second measurement is ignored"
                            )
                        else:
                            measured_wires.add(index)

            if measured_wires and len(measured_wires) != len(circuit.qubits):
                warnings.warn(
                    f"Number of wires in the circuit ({len(circuit.qubits)}) does not equal the "
                    f"number of wires with measurement instructions ({len(measured_wires)}). "
                    f"{self.__class__.__name__} only supports measurement of the entire quantum "
                    "register which will be performed instead."
                )

            if not measured_wires:
                solver.shots = None

            simulation_result = solver(circuit)

            output["results"].append(
                {
                    "header": {"name": exp_name, "random_seed": seed},
                    "shots": shots,
                    "spin": spin,
                    "status": "DONE",
                    "success": True,
                }
            )
            # add the simulation result at the correct place in the result dictionary
            output["results"][exp_i]["data"] = simulation_result

        output["job_id"] = job_id
        output["date"] = datetime.datetime.now().isoformat()
        output["backend_name"] = self.name()
        output["backend_version"] = self.configuration().backend_version
        output["time_taken"] = time.time() - start
        output["success"] = True
        output["qobj_id"] = None

        return Result.from_dict(output)

    # pylint: disable=arguments-differ, unused-argument
    def run(
        self,
        circuits: Union[QuantumCircuit, List[QuantumCircuit]],
        shots: int = 1000,
        spin: Union[float, Fraction] = Fraction(1, 2),
        seed: Optional[int] = None,
        **run_kwargs,
    ) -> AerJob:
        """
        Run the simulator with a variable length of the individual spins.

        Args:
            circuits: A list of quantum circuits.
            shots: The number of shots to measure.
            spin: The spin length of the simulated system which must be a positive
                integer or half-integer. Defaults to 1/2 which is equivalent to qubits.
            seed: The seed for the simulator.
            run_kwargs: Additional keyword arguments that might be passed down when calling
            qiskit.execute() which will have no effect on this backend.

        Returns:
             aer_job: a job object containing the result of the simulation
        """

        if isinstance(circuits, QuantumCircuit):
            circuits = [circuits]

        data = {"spin": spin, "shots": shots, "seed": seed, "experiments": {}}

        for idx, circuit in enumerate(circuits):
            data["experiments"]["experiment_%i" % idx] = {
                "circuit": circuit,
            }

        job_id = str(uuid.uuid4())
        aer_job = AerJob(self, job_id, self._execute, data)
        aer_job.submit()
        return aer_job



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Gates for spin backends."""

from typing import Union, Optional, List
from fractions import Fraction
import numpy as np
from scipy.linalg import expm

from qiskit.circuit.gate import Instruction, Gate
from qiskit_nature.operators.second_quantization import SpinOp
from qiskit_cold_atom import QiskitColdAtomError, add_gate


class SpinGate(Gate):
    """Unitary gates for spin circuits."""

    def __init__(
        self,
        name: str,
        num_modes: int,
        params: Optional[List] = None,
        label: Optional[str] = None,
        generator: Optional[SpinOp] = None,
    ) -> None:
        """Create a new spin gate.

        Args:
            name: The Qobj name of the gate.
            num_modes: The number of fermionic modes the gate acts on.
            params: A list of parameters.
            label: An optional label for the gate.
            generator: The generating Hamiltonian of the gate unitary given as a SpinOp
        """

        self._generator = generator

        self.num_modes = num_modes

        if params is None:
            params = []

        super().__init__(name=name, num_qubits=num_modes, params=params, label=label)

    def power(self, exponent: float):
        """Creates a spin gate as `gate^exponent`.

        Args:
            exponent (float): The exponent with which the gate is exponentiated

        Returns:
            SpinGate: To which `.generator` is self.generator*exponent.

        Raises:
            QiskitColdAtomError: Ff the gate generator is not defined.
        """
        if self.generator is None:
            raise QiskitColdAtomError(
                "Gate can not be exponentiated if the gate generator is not defined."
            )
        # the generator of the exponentiated gate is the old generator times the exponent
        exp_generator = exponent * self.generator

        exp_params = None if not self.params else [exponent * param for param in self.params]

        exp_label = None if not self.label else self.label + f"^{exponent}"

        return SpinGate(
            name=self.name + f"^{exponent}",
            num_modes=self.num_modes,
            params=exp_params,
            label=exp_label,
            generator=exp_generator,
        )

    # pylint: disable=arguments-differ
    def to_matrix(self, spin: Union[float, Fraction] = Fraction(1, 2)) -> np.ndarray:
        """Return a Numpy.array for the gate unitary matrix.

        Args:
            spin: The spin value of each wire that the gate acts on

        Returns:
            A dense np.array of the unitary of the gate
        """
        spin_op = SpinOp(self.generator.to_list(), spin=spin, register_length=self.num_qubits)
        return expm(-1j * spin_op.to_matrix())

    def control(
        self,
        num_ctrl_qubits: Optional[int] = 1,
        label: Optional[str] = None,
        ctrl_state: Optional[Union[int, str]] = None,
    ):
        """Overwrite control method which is supposed to return a controlled version of the gate.
        This is not applicable in the spin setting."""
        raise QiskitColdAtomError("Spin gates have no controlled version")

    @property
    def generator(self) -> SpinOp:
        """The Hamiltonian that generates the unitary of the gate, given as a SpinOp."""
        return self._generator


class RLXGate(SpinGate):
    r"""Rotation of the collective spin of a cold atomic Bose-Einstein
    condensate around the x-axis.

    The generating Hamiltonian of the LX gate is

    :math:`H = \omega L_x`

    where :math:`\omega` is the free gate parameter.

    **Circuit symbol:**

    .. parsed-literal::

             ┌────────────┐
        q_0: ┤ RLX(omega) ├
             └────────────┘
    """

    def __init__(self, omega, label=None):
        """Create new RLX gate."""
        super().__init__("rlx", 1, [omega], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the LX gate."""
        return float(self.params[0]) * SpinOp("X")


@add_gate
def rlx(self, omega, wire):
    # pylint: disable=invalid-name
    """add the RLX gate to a QuantumCircuit"""
    return self.append(RLXGate(omega), [wire], [])


class RLYGate(SpinGate):
    r"""Rotation of the collective spin of a cold atomic Bose-Einstein
    condensate around the y-axis.

    The generating Hamiltonian of the LY gate is

    :math:`H = \omega L_y`

    where :math:`\omega` is the free gate parameter

    **Circuit symbol:**

    .. parsed-literal::

             ┌────────────┐
        q_0: ┤ RLY(omega) ├
             └────────────┘
    """

    def __init__(self, omega, label=None):
        """Create new RLY gate."""
        super().__init__("rly", 1, [omega], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the LY gate."""
        return float(self.params[0]) * SpinOp("Y")


@add_gate
def rly(self, omega, wire):
    # pylint: disable=invalid-name
    """add the RLY gate to a QuantumCircuit"""
    return self.append(RLYGate(omega), [wire], [])


class RLZGate(SpinGate):
    r"""Rotation of the collective spin of a cold atomic Bose-Einstein condensate around the z-axis.

    The generating Hamiltonian of the LZ gate is

    :math:`H = \delta L_z`

    where :math:`\delta` is the free gate parameter

    **Circuit symbol:**

    .. parsed-literal::

             ┌────────────┐
        q_0: ┤ RLZ(delta) ├
             └────────────┘

    """

    def __init__(self, delta, label=None):
        """Create new RLZ gate."""
        super().__init__("rlz", 1, [delta], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the LZ gate."""
        return float(self.params[0]) * SpinOp("Z")


@add_gate
def rlz(self, delta, wire):
    # pylint: disable=invalid-name
    """add the RLZ gate to a QuantumCircuit"""
    return self.append(RLZGate(delta), [wire], [])


class RLZ2Gate(SpinGate):
    r"""Evolution of a coherent spin under the twisting dynamic generated by Lz^2'.

    The generating Hamiltonian of the LZ2 gate is

    :math:`H = \chi L_z^2`

    where :math:`\chi` is the free gate parameter.

    **Circuit symbol:**

    .. parsed-literal::

             ┌───────────┐
        q_0: ┤ RLZ2(chi) ├
             └───────────┘
    """

    def __init__(self, chi, label=None):
        """Create new rz2 gate."""
        super().__init__("rlz2", 1, [chi], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the LZ gate."""
        return float(self.params[0]) * SpinOp("Z_0^2", register_length=1)


@add_gate
def rlz2(self, chi, wire):
    # pylint: disable=invalid-name
    """Add the RLZ2 gate to a QuantumCircuit."""
    return self.append(RLZ2Gate(chi), [wire], [])


class OATGate(SpinGate):
    r"""Evolution of a coherent spin under the one-axis-twisting Hamiltonian.

    The generating Hamiltonian of the OATgate is

    :math:`H = \chi L_z^2 + \Delta L_z + \Omega L_x`

    where :math:`\chi`, :math:`\Delta` and :math:`\Omega` are the free gate parameters.
    """

    def __init__(self, chi: float, delta: float, omega: float, label=None):
        """Create new one-axis-twisting rotation gate."""
        super().__init__("OAT", 1, [chi, delta, omega], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the OAT gate."""
        return (
            float(self.params[0]) * SpinOp("Z_0^2", register_length=1)
            + float(self.params[1]) * SpinOp("Z")
            + float(self.params[2]) * SpinOp("X")
        )


@add_gate
def oat(self, chi: float, delta: float, omega: float, wire: int, label=None):
    """Add the RLZ2 gate to a QuantumCircuit."""
    return self.append(OATGate(chi=chi, delta=delta, omega=omega, label=label), [wire], [])


class RLZLZGate(SpinGate):
    r"""Coupled ZZ-rotation of two collective spins.

    The generating Hamiltonian of the RLZLZGate is

    :math:`H = \gamma L_{z, i} + L_{z, j}`

    where :math:`\gamma` is the free gate parameter and :math:`i` and :math:`j` index the wires
    the gate acts on.
    """

    def __init__(self, gamma: float, label=None):
        """Create new RLZLZ gate."""
        super().__init__("rlzlz", 2, [gamma], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the LZZ gate."""
        return self.params[0] * SpinOp("Z_0 Z_1", register_length=2)


@add_gate
def rlzlz(self, gamma: float, wires: List[int], label=None):
    """Add the RLZLZ gate to a QuantumCircuit."""
    return self.append(RLZLZGate(gamma=gamma, label=label), qargs=wires)


class RLXLYGate(SpinGate):
    r"""The spin exchange gate of two collective spins.

    The generating Hamiltonian of the LxLyGate is

    :math:`H = \gamma (L_{x, i}L_{x,j} + L_{y, i} L_{y,j})`

    where :math:`\gamma` is the free gate parameter and :math:`i` and :math:`j` index the wires the gate
    acts on. This gate is equivalently expressed through raising and lowering operators as:

    :math:`H = 2\gamma (L_{+, i}L_{-,j} + L_{-, i} L_{+,j})`
    """

    def __init__(self, gamma: float, label=None):
        """Create new RLXLY gate."""
        super().__init__("rlxly", 2, [gamma], label=label)

    @property
    def generator(self) -> SpinOp:
        r"""The generating Hamiltonian of the LxLy gate."""
        return self.params[0] * (
            SpinOp("X_0 X_1", register_length=2) + SpinOp("Y_0 Y_1", register_length=2)
        )


@add_gate
def rlxly(self, gamma: float, wires: List[int], label=None):
    """Add the RLXLY gate to a QuantumCircuit."""
    return self.append(RLXLYGate(gamma=gamma, label=label), qargs=wires)


class LoadSpins(Instruction):
    r"""An instruction to define the spin length of each qudit mode.

    This gate loads `num_atoms` onto the wire of index `wire`. This results in a
    local spin length of :math:`\ell = N/2`.

    **Circuit symbol:**

    .. parsed-literal::

             ┌──────┐
        q_0: ┤ Load ├
             └──────┘
    """

    def __init__(self, num_atoms: int):
        """Initialise new load instruction.

        Args:
            num_atoms: The integer number of atoms loaded into this wire. n Qobj name of the gate.
        """
        super().__init__(name="load", num_qubits=1, num_clbits=0, params=[num_atoms], label=None)


@add_gate
def load_spins(self, num_atoms: int, wire: int):
    """Add the load spin gate to a QuantumCircuit."""
    return self.append(LoadSpins(num_atoms), [wire], [])


class RydbergFull(SpinGate):
    r"""
    Global 1D-Rydberg dynamic consisting of the detuning, Rabi coupling and Rydberg blockade.

    The generating Hamiltonian of the Rydberg dynamcis is

    :math:` \hat{H}_{ryd} = \Omega \sum_i \sigma_{x,i}
    + \Delta \sum_i \sigma_{z,i}
    + \phi \sum_{i\neq j} \frac{\hat{n}_i \hat{n}_j}{|i-j|^6}`

    where :math:`i` indexes the mode, :math:`\sigma` indexes the spin,
    :math:`\Omega` are the Rabi couplings, :math:`\phi` is the interaction strength and
    :math:`\Delta` are the detunings. The form of the third part of the Hamiltonian
    implies that the atoms are arranged in a regularly spaced 1D chain.

    **Circuit symbol:**

        .. parsed-literal::

                 ┌────────────────┐
            q_0: ┤   RydbergFull  ├
                 └────────────────┘

    """

    def __init__(self, num_modes: int, omega: float, delta: float, phi: float, label=None):
        """Initialize a global Rydberg gate

        Args:
            num_modes: number of tweezers on which the operator acts, must be entire quantum register
            omega: global strength of the Rabi coupling on each site.
            delta: global detuning
            phi: global interaction strength
            label: optional
        """

        params = [omega, delta, phi]

        super().__init__(
            name="rydberg_full",
            num_modes=num_modes,
            params=params,
            label=label,
        )

    def inverse(self):
        """Get the inverse gate by reversing the sign of all gate parameters"""
        omega_val, delta_val, phi_val = self.params[0], self.params[1], self.params[2]

        return RydbergFull(
            num_modes=self.num_modes, omega=-omega_val, delta=-delta_val, phi=-phi_val
        )

    @property
    def generator(self) -> SpinOp:
        """The generating Hamiltonian of the Rydberg Hamiltonian."""
        params = [float(param) for param in self.params]
        omega, delta, phi = params[0], params[1], params[2]
        generators = []

        # add generators of Rabi coupling
        if omega != 0.0:
            for i in range(self.num_modes):
                generators.append((f"X_{i}", omega))
        # add generators of detuning
        if delta != 0.0:
            for i in range(self.num_modes):
                generators.append((f"Z_{i}", delta))

        # add generators of interaction term
        if phi != 0.0:
            for i in range(self.num_modes):
                for j in range(i + 1, self.num_modes):
                    coeff = phi / np.abs(i - j) ** 6
                    generators.append((f"Z_{i} Z_{j}", coeff))
                    generators.append((f"Z_{i}", -coeff / 2))
                    generators.append((f"Z_{j}", -coeff / 2))

        if not generators:
            return SpinOp("I_0", register_length=self.num_modes)
        else:
            return sum(
                coeff * SpinOp(label, register_length=self.num_modes) for label, coeff in generators
            )


# pylint: disable=invalid-name
@add_gate
def rydberg_full(self, omega: float, delta: float, phi: float, modes: List[int], label=None):
    """Add the combined Rydberg Gate gate to a QuantumCircuit."""
    return self.append(
        RydbergFull(num_modes=len(modes), omega=omega, delta=delta, phi=phi, label=label),
        qargs=modes,
    )


class RydbergBlockade(SpinGate):
    r"""
    Global 1D-Rydberg blockade.

    The generating Hamiltonian of the Rydberg blockade is

    :math:`\hat{H}_{block} =  \sum_{i\neq j} \frac{\hat{n}_i \hat{n}_j}{|i-j|^6}`

    where :math:`i` indexes the mode, :math:`\sigma` indexes the spin. The form of the Hamiltonian
    implies that the atoms are arranged in a regularly spaced 1D chain.

    **Circuit symbol:**

        .. parsed-literal::

                 ┌────────────────────┐
            q_0: ┤   RydbergBlockade  ├
                 └────────────────────┘

    """

    def __init__(self, num_modes: int, phi: float, label=None):
        """Initialize a global Rydberg gate

        Args:
            num_modes: number of tweezers on which the operator acts, must be entire quantum register
            phi: global interaction strength
            label: optional
        """
        super().__init__(
            name="rydberg_blockade",
            num_modes=num_modes,
            params=[phi],
            label=label,
        )

    def inverse(self):
        """Get the inverse gate by reversing the sign of all gate parameters"""
        return RydbergBlockade(num_modes=self.num_modes, phi=-self.params[0])

    @property
    def generator(self) -> SpinOp:
        """The generating Hamiltonian of the Rydberg Blockade Hamiltonian."""
        phi = float(self.params[0])
        generators = []

        # add generators of interaction term
        if phi != 0.0:
            for i in range(self.num_modes):
                for j in range(i + 1, self.num_modes):
                    coeff = phi / np.abs(i - j) ** 6
                    generators.append((f"Z_{i} Z_{j}", coeff))
                    generators.append((f"Z_{i}", -coeff / 2))
                    generators.append((f"Z_{j}", -coeff / 2))

        if not generators:
            return SpinOp("I_0", register_length=self.num_modes)
        else:
            return sum(
                coeff * SpinOp(label, register_length=self.num_modes) for label, coeff in generators
            )


# pylint: disable=invalid-name
@add_gate
def rydberg_block(self, phi: float, modes: List[int], label=None):
    """Add the  Rydberg Blockade gate to a QuantumCircuit."""
    return self.append(RydbergBlockade(num_modes=len(modes), phi=phi, label=label), qargs=modes)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""General spin simulator backend tests."""

from time import sleep
import numpy as np

from qiskit import QuantumCircuit
from qiskit.providers import JobStatus
from qiskit.result import Result
from qiskit.test import QiskitTestCase
from qiskit_aer import AerJob

from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom.spins import SpinSimulator
from qiskit_cold_atom.spins.base_spin_backend import BaseSpinBackend


class TestSpinSimulatorBackend(QiskitTestCase):
    """class to test the FermionSimulatorBackend class."""

    def setUp(self):
        super().setUp()
        self.backend = SpinSimulator()

    def test_initialization(self):
        """test the initialization of the backend"""
        target_config = {
            "backend_name": "spin_simulator",
            "backend_version": "0.0.1",
            "n_qubits": None,
            "basis_gates": None,
            "gates": [],
            "local": True,
            "simulator": True,
            "conditional": False,
            "open_pulse": False,
            "memory": True,
            "max_shots": 1e5,
            "coupling_map": None,
            "description": "a base simulator for spin circuits. Instead of a qubit, each wire "
            "represents a single high-dimensional spin",
        }

        backend = SpinSimulator()
        self.assertIsInstance(backend, BaseSpinBackend)
        self.assertTrue(target_config.items() <= backend.configuration().to_dict().items())

    def test_run_method(self):
        """Test the run method of the backend simulator"""

        with self.subTest("test call"):
            circ = QuantumCircuit(2)
            job = self.backend.run(circ)
            self.assertIsInstance(job, AerJob)
            self.assertIsInstance(job.job_id(), str)
            self.assertIsInstance(job.result(), Result)
            self.assertEqual(job.status(), JobStatus.DONE)

        circ1 = QuantumCircuit(2)
        circ2 = QuantumCircuit(3)

        with self.subTest("test call with multiple circuits"):
            job = self.backend.run([circ1, circ2])
            self.assertIsInstance(job, AerJob)

        with self.subTest("test shot number"):
            target_shots = 123
            job = self.backend.run([circ1, circ2], shots=target_shots)
            for exp in job.result().results:
                self.assertEqual(exp.shots, target_shots)

        with self.subTest("test seed of RNG"):
            target_seed = 123
            job = self.backend.run([circ1, circ2], seed=target_seed)
            for exp in job.result().results:
                self.assertEqual(exp.header.random_seed, target_seed)

        with self.subTest("test dimension of simulation"):
            test_circ = QuantumCircuit(2)
            test_circ.rlx(np.pi / 2, 0)
            test_circ.rly(np.pi / 4, [0, 1])

            statevector_1 = self.backend.run(test_circ, spin=1).result().get_statevector()
            self.assertEqual(len(statevector_1), 3**2)

            statevector_2 = self.backend.run(test_circ, spin=5 / 2).result().get_statevector()
            self.assertEqual(len(statevector_2), 6**2)

        with self.subTest("test irregular spin values"):
            test_circ = QuantumCircuit(2)
            job = self.backend.run(test_circ, spin=5 / 4)
            sleep(0.01)
            self.assertIs(job.status(), JobStatus.ERROR)
            with self.assertRaises(QiskitColdAtomError):
                job.result()

    def test_execute(self):
        """test the ._execute() method internally called by .run()"""

        test_circ = QuantumCircuit(2)
        test_circ.rly(np.pi / 2, 0)
        test_circ.rlx(np.pi / 2, 1)
        test_circ.measure_all()

        result = self.backend.run(test_circ, spin=1, seed=45, shots=5).result()

        with self.subTest("test simulation counts"):
            self.assertEqual(result.get_counts(), {"0 1": 1, "2 2": 1, "1 1": 2, "1 0": 1})

        with self.subTest("test simulation memory"):
            self.assertEqual(result.get_memory(), ["2 2", "1 1", "0 1", "1 0", "1 1"])

        with self.subTest("test simulation statevector"):
            self.assertTrue(
                np.allclose(
                    result.get_statevector(),
                    np.array(
                        [
                            1 / 4,
                            -1j / np.sqrt(8),
                            -1 / 4,
                            1 / np.sqrt(8),
                            -1j / 2,
                            -1 / np.sqrt(8),
                            1 / 4,
                            -1j / np.sqrt(8),
                            -1 / 4,
                        ]
                    ),
                )
            )

        with self.subTest("test simulation unitary"):
            # test the unitary on a single spin-2 example
            test_circ = QuantumCircuit(1)
            test_circ.rlx(np.pi / 2, 0)
            test_circ.rlz(np.pi / 2, 0)
            test_circ.measure_all()

            result = self.backend.run(test_circ, spin=2, seed=45, shots=5).result()

            self.assertTrue(
                np.allclose(
                    result.get_unitary(),
                    np.array(
                        [
                            [-0.25, 0.5j, np.sqrt(3 / 8), -0.5j, -0.25],
                            [-0.5, 0.5j, 0.0, 0.5j, 0.5],
                            [-np.sqrt(3 / 8), 0.0, -0.5, 0.0, -np.sqrt(3 / 8)],
                            [-0.5, -0.5j, 0.0, -0.5j, 0.5],
                            [-0.25, -0.5j, np.sqrt(3 / 8), 0.5j, -0.25],
                        ]
                    ),
                )
            )

        with self.subTest("test time taken"):
            self.assertTrue(result.to_dict()["time_taken"] < 0.5)

        with self.subTest("test result success"):
            self.assertTrue(result.to_dict()["success"])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Provider for cold atomic experiments."""

import os
import warnings
import json
from configparser import ConfigParser, ParsingError
from typing import Callable, Dict, Optional, Union, List
import errno

import requests

from qiskit.providers import ProviderV1 as Provider
from qiskit.providers import BackendV1 as Backend
from qiskit.providers.providerutils import filter_backends
from qiskit.providers.exceptions import QiskitBackendNotFoundError
from qiskit_cold_atom.exceptions import QiskitColdAtomError

from qiskit_cold_atom.fermions import FermionSimulator
from qiskit_cold_atom.spins import SpinSimulator

from qiskit_cold_atom.providers.fermionic_tweezer_backend import (
    FermionicTweezerSimulator,
)
from qiskit_cold_atom.providers.remote_backend import (
    RemoteSpinBackend,
    RemoteFermionBackend,
)
from qiskit_cold_atom.providers.collective_spin_backend import CollectiveSpinSimulator

# Default location of the credentials file
_DEFAULT_COLDATOM_FILE = os.path.join(
    os.path.expanduser("~"), ".qiskit", "cold_atom_credentials.conf"
)


class ColdAtomProvider(Provider):
    """
    Provider for cold atomic backends. To access the devices, the user has to specify the url of the
    backend, a username and an access token.
    With these credentials, the provider can be accessed like this:

    .. code-block:: python

        from qiskit_cold_atom.providers import ColdAtomProvider
        # save account with personal credentials
        ColdAtomProvider.save_account(url="my_url", username="my_name", token="my_url")
        # load account
        provider = ColdAtomProvider.load_account()
        # access backends
        backends = provider.backends()
    """

    def __init__(self, credentials: Optional[Dict[str, Union[str, List[str]]]] = None):
        """
        Constructor of the cold atom provider. Using the given credentials, the provider tries to
        access the remote backends and adds all backends that can be reached to the available backends.

        Args:
            credentials: A dictionary of the shape
            {'urls' = ['http://...', ...] 'username' = '...', 'token' = '...'}
            The value to the url key could also be a single url

        Raises:
            QiskitColdAtomError: If the credentials file specified does not match the expected format
        """

        self.credentials = credentials
        self.name = "qiskit_cold_atom_provider"

        # Populate the list of backends with the local simulators
        backends = [
            FermionSimulator(provider=self),
            SpinSimulator(provider=self),
            FermionicTweezerSimulator(provider=self),
            CollectiveSpinSimulator(provider=self),
        ]

        if credentials is not None:

            try:
                urls = self.credentials["urls"]
                name = self.credentials["username"]
                token = self.credentials["token"]
            except KeyError as key_err:
                raise QiskitColdAtomError(
                    "Credentials do not match the expected format"
                ) from key_err
            if isinstance(urls, str):
                urls = [urls]

            # try to ping remote backend with credentials:
            for url in urls:
                try:
                    r = requests.get(
                        url + "/get_config",
                        params={"username": name, "token": token},
                    )

                    backend_type = r.json()["cold_atom_type"]

                    if backend_type == "fermion":
                        backend = RemoteFermionBackend
                    elif backend_type == "spin":
                        backend = RemoteSpinBackend
                    else:
                        warnings.warn(
                            f"specified cold atom type {backend_type} at url {url} is not recognized"
                        )
                        continue

                    backends.append(backend(provider=self, url=url))

                except requests.exceptions.RequestException as request_error:
                    warnings.warn(
                        f"connection to the remote backend could not be established: {request_error}"
                    )
                except json.JSONDecodeError as json_error:
                    warnings.warn(
                        f"Backend configuration could not be read: {r.text}, {json_error}"
                    )
                except KeyError:
                    warnings.warn(f"backend at url {url} has no specified cold atom type")

        self._backends = BackendService(backends)

    def __str__(self):
        return f"<{self.__class__.__name__}(name={self.name})>"

    def __repr__(self):
        return self.__str__()

    def get_backend(self, name: str = None, **kwargs) -> Backend:
        """Return a single backend matching the specified filtering.
        Args:
            name: name of the backend.
            **kwargs: dict used for filtering.
        Returns:
            backend: a backend matching the filtering.
        Raises:
            QiskitBackendNotFoundError: if no backend could be found or
                more than one backend matches the filtering criteria.
        """
        backends = self._backends(name, **kwargs)
        if len(backends) > 1:
            raise QiskitBackendNotFoundError("More than one backend matches criteria.")
        if not backends:
            raise QiskitBackendNotFoundError("No backend matches criteria.")

        return backends[0]

    @staticmethod
    def save_account(
        url: Union[str, List[str]],
        username: str,
        token: str,
        overwrite: bool = False,
        filename: Optional[str] = None,
    ) -> None:
        """Save credentials to a file locally

        Args:
            url: The url(s) of the backends(s) to connect to
            username: username of the account
            token: the token of the account
            overwrite: If true, will overwrite any credentials already stored on disk
            filename: Full path to the credentials file. If ``None``, the default
                location is used (``$HOME/.qiskit/cold_atom_credentials``).

        Raises:
            OSError: If there is a race condition when creating the directory for the
                credentials if it does not already exist.
        """
        if isinstance(url, str):
            url = [url]

        filename = filename or _DEFAULT_COLDATOM_FILE

        credentials_present = False

        if os.path.isfile(filename):

            stored_credentials = ColdAtomProvider.stored_account(filename=filename)

            if stored_credentials != {}:
                credentials_present = True

            if credentials_present and not overwrite:
                warnings.warn("Credentials already present. Set overwrite=True to overwrite.")
        else:
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        if not credentials_present or overwrite:

            credentials = {
                "cold-atom-credentials": {
                    "urls": " ".join(url),
                    "username": username,
                    "token": token,
                }
            }
            # Write the credentials file.
            with open(filename, "w") as credentials_file:
                config_parser = ConfigParser()
                config_parser.read_dict(credentials)
                config_parser.write(credentials_file)

    @staticmethod
    def add_url(
        url: str,
        filename: Optional[str] = None,
    ) -> None:
        """Add a url to the existing credentials stored on disk

        Args:
            url: The url of a backend to add to the existing credentials
            filename: Full path to the credentials file. If ``None``, the default
                location is used (``$HOME/.qiskit/cold_atom_credentials``).

        Raises:
            QiskitColdAtomError: If there are no stored credentials at the specified location
        """
        credentials_dict = ColdAtomProvider.stored_account(filename)
        try:
            urls = credentials_dict["urls"]
            name = credentials_dict["username"]
            token = credentials_dict["token"]
        except KeyError as err:
            raise QiskitColdAtomError("No stored credentials found") from err
        urls.append(url)
        ColdAtomProvider.save_account(urls, name, token, overwrite=True, filename=filename)

    @classmethod
    def load_account(cls, filename: Optional[str] = None) -> "ColdAtomProvider":
        """Check for stored credentials and return a provider instance. If no credentials are found
        the provider will not include the remote backends bu only the simulators

        Args:
            filename: Full path to the credentials file. If ``None``, the default
                location is used (``$HOME/.qiskit/cold_atom_credentials``).

        Returns:
            A provider instance initialized with the backends available to the account
        """

        stored_credentials = ColdAtomProvider.stored_account(filename=filename)
        if stored_credentials == {}:
            warnings.warn("No stored credentials found")
            return cls()
        else:
            return cls(stored_credentials)

    @classmethod
    def enable_account(
        cls, url: Union[str, List[str]], username: str, token: str
    ) -> "ColdAtomProvider":
        """Create provider from credentials given at runtime without storing credentials

        Args:
            url: The url(s) of the backends(s) to connect to
            username: username of the account
            token: the token of the account

        Returns:
            A provider instance initialized with the backends available to the account
        """
        if isinstance(url, str):
            url = [url]
        credentials = {"urls": url, "username": username, "token": token}
        return cls(credentials)

    @staticmethod
    def delete_account(filename: Optional[str] = None) -> None:
        """Delete any credentials saved locally at filename. This will not delete the file itself but
        erase the content of that file

        Args:
            filename: Full path to the credentials file. If ``None``, the default
                location is used (``$HOME/.qiskit/cold_atom_credentials``).
        """
        filename = filename or _DEFAULT_COLDATOM_FILE

        if not os.path.isfile(filename):
            warnings.warn("No credentials found at the specified location")

        else:
            # Write an empty dictionary to the file
            with open(filename, "w") as credentials_file:
                config_parser = ConfigParser()
                config_parser.read_dict({})
                config_parser.write(credentials_file)

    @staticmethod
    def stored_account(
        filename: Optional[str] = None,
    ) -> Dict[str, Union[str, List[str]]]:
        """Retrieve the credentials stored on disk.

        Args:
            filename: Full path to the credentials file. If ``None``, the default
                location is used (``$HOME/.qiskit/cold_atom_credentials``).

        Raises:
            QiskitColdAtomError: If the file specified does not match the expected format

        Returns:
            A dictionary containing the found credentials. This dictionary is empty if no credentials
            are found at the specified locations
        """
        filename = filename or _DEFAULT_COLDATOM_FILE
        config_parser = ConfigParser()
        try:
            config_parser.read(filename)
        except ParsingError as ex:
            raise QiskitColdAtomError(f"Error parsing file {filename}: {str(ex)}") from ex

        credentials_dict = {}

        for name in config_parser.sections():
            if name.startswith("cold-atom-credentials"):
                credentials_dict = dict(config_parser.items(name))
                credentials_dict["urls"] = credentials_dict["urls"].split()
            else:
                raise QiskitColdAtomError(f"stored credentials file has unknown section {name}")

        return credentials_dict

    def active_account(self) -> Optional[Dict[str, str]]:
        """Return the credentials in use for the session."""
        return self.credentials

    # pylint: disable=arguments-differ
    def backends(self, name: Optional[str] = None, filters: Optional[Callable] = None, **kwargs):
        """Abstract method of the Base Provider class"""
        return self._backends(name, filters, **kwargs)


class BackendService:
    """A service class that allows for autocompletion
    of backends from provider.
    """

    def __init__(self, backends):
        """Initialize service
        Parameters:
            backends (list): List of backend instances.
        """
        self._backends = backends
        for backend in backends:
            setattr(self, backend.name(), backend)

    def __call__(self, name: str = None, filters: Callable = None, **kwargs):
        """A listing of all backends from this provider.
        Parameters:
            name: The name of a given backend.
            filters: A filter function.
        Returns:
            list: A list of backends, if any.
        """
        # pylint: disable=arguments-differ
        backends = self._backends
        if name:
            backends = [backend for backend in backends if backend.name() == name]

        return filter_backends(backends, filters=filters, **kwargs)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Cold Atom Backend Provider
===========================

This module contains a provider to interface with cold atomic device backends.
To access these devices, the user has to specify the url of the backend, a username and an access token.
With these credentials, the provider can be accessed like this:

.. code-block:: python

    from qiskit_cold_atom.providers import ColdAtomProvider

    # save account with personal credentials
    ColdAtomProvider.save_account(url="my_url", username="my_name", token="my_url")

    provider = ColdAtomProvider.load_account()
    backends = provider.backends()

.. autosummary::
   :toctree: ../stubs/

    ColdAtomProvider


Cold Atom Jobs
==============

When executing jobs, cold atom backends will return instances of :class:`ColdAtomJob`.

.. autosummary::
   :toctree: ../stubs/

    ColdAtomJob

"""

from .cold_atom_provider import ColdAtomProvider
from .cold_atom_job import ColdAtomJob



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module to describe fermionic states in occupation number basis"""

from typing import List, Union
import warnings
import numpy as np

from qiskit import QuantumCircuit

from qiskit_cold_atom.exceptions import QiskitColdAtomError


class FermionicState:
    """Fermionic states in an occupation number representation."""

    def __init__(self, occupations: Union[List[int], List[List[int]]]):
        """Create a :class:`FermionicState` from the given occupations.

        Args:
            occupations: List of occupation numbers. When List[int] is given, the occupations
            correspond to the number of indistinguishable fermionic particles in each mode,
            e.g. [0, 1, 1, 0] implies that sites one and two are occupied by a fermion.
            When List[List[int]] is given, the occupations describe the number of particles in
            fermionic modes with different (distinguishable) species of fermions. Each
            inner list gives the occupations of one fermionic species.

        Raises:
            QiskitColdAtomError:
                - If the inner lists do not have the same length
                - If the occupations are not 0 or 1
        """

        if isinstance(occupations[0], int):
            occupations = [occupations]

        self._sites = len(occupations[0])
        self._occupations = occupations
        self._num_species = len(occupations)

        self._occupations_flat = []
        for occs in self.occupations:
            self._occupations_flat += occs

        for occs in self.occupations[0:]:
            if len(occs) != self._sites:
                raise QiskitColdAtomError(
                    f"All occupations of different fermionic species must have "
                    f"same length, received {self.occupations[0]} and {occs}."
                )
            for n in occs:
                if n not in (0, 1):
                    raise QiskitColdAtomError(f"Fermionic occupations must be 0 or 1, got {n}.")

    @property
    def occupations(self) -> List[List[int]]:
        """Return the occupation number of each fermionic mode."""
        return self._occupations

    @property
    def occupations_flat(self) -> List[int]:
        """Return the occupations of each fermionic mode in a flat list."""
        return self._occupations_flat

    @property
    def sites(self) -> int:
        """Return the number of fermionic sites."""
        return self._sites

    @property
    def num_species(self) -> int:
        """Return the number of species of fermions, e.g. 2 for spin up/down systems."""
        return self._num_species

    def __str__(self):
        output = ""
        for i in range(self.num_species):
            output += "|" + str(self.occupations[i])[1:-1] + ">"
        return output

    @classmethod
    def from_total_occupations(cls, occupations: List[int], num_species: int) -> "FermionicState":
        """
        Create a fermionic state from a single (flat) list of total occupations.

        Args:
            occupations: a list of occupations of all fermionic modes, e.g. [0, 1, 1, 0, 1, 0].
            num_species: number of fermionic species. If > 1, the total occupation list is cast
                into a nested list where each inner list describes one fermionic species. In the
                above example, for num_species = 2, this becomes
                FermionicState([[0, 1, 1], [0, 1, 0]]).

        Returns:
            A fermionic state initialized with the given input.

        Raises:
            QiskitColdAtomError: If the length of occupations is not a multiple of num_species.
        """
        if len(occupations) % num_species != 0:
            raise QiskitColdAtomError(
                "The state must have a number of occupations that is a multiple of the"
                "number of fermionic species."
            )

        sites = int(len(occupations) / num_species)
        return cls(np.reshape(occupations, (num_species, sites)).tolist())

    @classmethod
    def initial_state(cls, circuit: QuantumCircuit, num_species: int = 1) -> "FermionicState":
        """
        Create a fermionic state from a quantum circuit that uses the `LoadFermion` instruction.
        This instruction must be the first instructions of the circuit and no further LoadFermion
        instruction can be applied, even after other instructions such as gates have been applied.

        Args:
            circuit: a quantum circuit with LoadFermions instructions that initialize fermionic
                particles.
            num_species: number of different fermionic species, e.g. 1 for a single
                type of spinless fermions (default), 2 for spin-1/2 fermions etc.

        Returns:
            A FermionicState initialized from the given circuit.

        Raises:
            QiskitColdAtomError:
                - If the number of wires in the circuit is not a multiple of num_species,
                - If LoadFermions instructions come after other instructions.
        """
        if num_species > 1:
            if circuit.num_qubits % num_species != 0:
                raise QiskitColdAtomError(
                    "The circuit must have a number of wires that is a multiple of the"
                    "number of fermionic species."
                )

        occupations = [0] * circuit.num_qubits
        gates_applied = [False] * circuit.num_qubits

        if not circuit.data[0][0].name == "load":
            warnings.warn(
                "No particles have been initialized, the circuit will return a trivial result."
            )

        # check that there are no more 'LoadFermions' instructions
        for instruction in circuit.data:

            qargs = [circuit.qubits.index(qubit) for qubit in instruction[1]]

            if instruction[0].name == "load":
                for idx in qargs:
                    if gates_applied[idx]:
                        raise QiskitColdAtomError(
                            f"State preparation instruction in circuit after gates on wire {idx}"
                        )
                    occupations[idx] = 1
            else:
                for idx in qargs:
                    gates_applied[idx] = True

        return cls.from_total_occupations(occupations, num_species)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""tests for circuit_to_cold_atom functions"""

from typing import Dict

from qiskit import QuantumCircuit
from qiskit.test import QiskitTestCase
from qiskit.circuit import Parameter
from qiskit.providers import BackendV1 as Backend
from qiskit.providers.models import BackendConfiguration
from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom.circuit_tools import CircuitTools, WireOrder

# These imports are needed to decorate the quantum circuit
import qiskit_cold_atom.spins  # pylint: disable=unused-import
import qiskit_cold_atom.fermions  # pylint: disable=unused-import


class DummyBackend(Backend):
    """dummy backend class for test purposes only"""

    def __init__(self, config_dict: Dict):
        super().__init__(configuration=BackendConfiguration.from_dict(config_dict))

    def run(self, run_input, **options):
        pass

    @classmethod
    def _default_options(cls):
        pass


class TestCircuitToColdAtom(QiskitTestCase):
    """circuit to cold atom tests."""

    def setUp(self):
        super().setUp()
        # Set up a dummy backend from a configuration dictionary

        test_config = {
            "backend_name": "test_backend",
            "backend_version": "0.0.1",
            "simulator": True,
            "local": True,
            "coupling_map": None,
            "description": "dummy backend for testing purposes only",
            "basis_gates": ["hop, int"],
            "memory": False,
            "n_qubits": 5,
            "conditional": False,
            "max_shots": 100,
            "max_experiments": 2,
            "open_pulse": False,
            "gates": [
                {
                    "coupling_map": [[0], [1], [2], [3], [4]],
                    "name": "rlz",
                    "parameters": ["delta"],
                    "qasm_def": "gate rLz(delta) {}",
                },
                {
                    "coupling_map": [[0], [1], [2]],
                    "name": "rlz2",
                    "parameters": ["chi"],
                    "qasm_def": "gate rlz2(chi) {}",
                },
                {
                    "coupling_map": [[0], [1], [2], [3], [4]],
                    "name": "rlx",
                    "parameters": ["omega"],
                    "qasm_def": "gate rx(omega) {}",
                },
            ],
            "supported_instructions": [
                "delay",
                "rlx",
                "rlz",
                "rlz2",
                "measure",
                "barrier",
            ],
        }

        self.dummy_backend = DummyBackend(test_config)

    def test_circuit_to_cold_atom(self):
        """test the circuit_to_cold_atom function"""

        circ1 = QuantumCircuit(3)
        circ1.rlx(0.5, [0, 1])
        circ1.rlz(0.3, [1, 2])
        circ1.measure_all()

        circ2 = QuantumCircuit(2)
        circ2.rlz2(0.5, 1)
        circ2.measure_all()

        shots = 10

        target_output = {
            "experiment_0": {
                "instructions": [
                    ["rlx", [0], [0.5]],
                    ["rlx", [1], [0.5]],
                    ["rlz", [1], [0.3]],
                    ["rlz", [2], [0.3]],
                    ["barrier", [0, 1, 2], []],
                    ["measure", [0], []],
                    ["measure", [1], []],
                    ["measure", [2], []],
                ],
                "num_wires": 3,
                "shots": shots,
                "wire_order": "sequential",
            },
            "experiment_1": {
                "instructions": [
                    ["rlz2", [1], [0.5]],
                    ["barrier", [0, 1], []],
                    ["measure", [0], []],
                    ["measure", [1], []],
                ],
                "num_wires": 2,
                "shots": shots,
                "wire_order": "sequential",
            },
        }

        actual_output = CircuitTools.circuit_to_cold_atom(
            [circ1, circ2], backend=self.dummy_backend, shots=shots
        )

        self.assertEqual(actual_output, target_output)

    def test_validate_circuits(self):
        """test the validation of circuits against the backend configuration"""

        with self.subTest("test size of circuit"):
            circ = QuantumCircuit(6)
            circ.rlx(0.4, 2)
            with self.assertRaises(QiskitColdAtomError):
                CircuitTools.validate_circuits(circ, backend=self.dummy_backend)

        with self.subTest("test support of native instructions"):
            circ = QuantumCircuit(4)
            # add gate that is not supported by the backend
            circ.fhop([0.5], [0, 1, 2, 3])
            with self.assertRaises(QiskitColdAtomError):
                CircuitTools.validate_circuits(circ, backend=self.dummy_backend)

        with self.subTest("check gate coupling map"):
            circ = QuantumCircuit(5)
            circ.rlz2(0.5, 4)
            with self.assertRaises(QiskitColdAtomError):
                CircuitTools.validate_circuits(circ, backend=self.dummy_backend)

        with self.subTest("test max. allowed circuits"):
            circuits = [QuantumCircuit(2)] * 3
            with self.assertRaises(QiskitColdAtomError):
                CircuitTools.circuit_to_cold_atom(circuits=circuits, backend=self.dummy_backend)

        with self.subTest("test max. allowed shots"):
            circuits = QuantumCircuit(2)
            with self.assertRaises(QiskitColdAtomError):
                CircuitTools.circuit_to_cold_atom(
                    circuits=circuits, backend=self.dummy_backend, shots=1000
                )

        with self.subTest("test running with unbound parameters"):
            theta = Parameter("θ")
            circ = QuantumCircuit(1)
            circ.rlx(theta, 0)
            with self.assertRaises(QiskitColdAtomError):
                CircuitTools.validate_circuits(circ, backend=self.dummy_backend)

    def test_circuit_to_data(self):
        """test the circuit to data method"""

        circ = QuantumCircuit(3)
        circ.rlx(0.5, [0, 1])
        circ.rlz(0.3, [1, 2])
        circ.measure_all()

        target_output = [
            ["rlx", [0], [0.5]],
            ["rlx", [1], [0.5]],
            ["rlz", [1], [0.3]],
            ["rlz", [2], [0.3]],
            ["barrier", [0, 1, 2], []],
            ["measure", [0], []],
            ["measure", [1], []],
            ["measure", [2], []],
        ]

        actual_output = CircuitTools.circuit_to_data(circ, backend=self.dummy_backend)

        self.assertEqual(actual_output, target_output)

    def test_convert_wire_order(self):
        """test the convert_wire_order method"""

        num_sites = 4
        num_species = 3
        # conversion rule: i -> (i % num_sites) * num_species + i // num_sites
        wires_sequential = [0, 1, 4, 5, 8, 9]
        wires_interleaved = [0, 3, 1, 4, 2, 5]

        with self.subTest("test sequential to interleaved"):
            wires_converted = CircuitTools.convert_wire_order(
                wires=wires_sequential,
                convention_from=WireOrder.SEQUENTIAL,
                convention_to=WireOrder.INTERLEAVED,
                num_sites=num_sites,
                num_species=num_species,
            )
            self.assertEqual(wires_converted, wires_interleaved)

        with self.subTest("test interleaved to sequential"):
            wires_converted = CircuitTools.convert_wire_order(
                wires=wires_interleaved,
                convention_from=WireOrder.INTERLEAVED,
                convention_to=WireOrder.SEQUENTIAL,
                num_sites=num_sites,
                num_species=num_species,
            )
            self.assertEqual(wires_converted, wires_sequential)
